/* See LICENSE file for copyright and license details. */
#include <dirent.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include "arg.h"
#include "base64.h"
#include "crypto_api.h"

#define SIGBEGIN "-----BEGIN ED25519 SIGNATURE-----\n"
#define SIGEND   "-----END ED25519 SIGNATURE-----\n"

enum {
	ACT_NONE,
	ACT_SIGN,
	ACT_CHCK,
	ACT_TRIM
};

enum {
	ERR_NOKEY  = 1,
	ERR_NOSIG  = 2,
	ERR_NOMSG  = 3,
	ERR_NORING = 4
};

static void usage();
static char *memstr(const void *h0, size_t k, const char *n0, size_t l);
static size_t bufferize(unsigned char **buf, FILE *fp);
static size_t extractmsg(unsigned char *msg[], unsigned char *buf, size_t len);
static size_t extractsig(unsigned char *sig[], unsigned char *buf, size_t len);
static int createkeypair(const char *);
static int readkey(FILE *, unsigned char **);
static int check_keyring(unsigned char *sig, unsigned char *msg, size_t len);
static int sign(FILE *fp, FILE *key);
static int check(FILE *fp, FILE *key);
static int trimsig(FILE *fp);

char *argv0;
static int verbose = 0;
static int armored = 0;

static void
usage()
{
	fprintf(stderr, "usage: %s [-astv] [-g ALIAS] [-f KEY] [FILE]\n",
			argv0);
	exit(EXIT_FAILURE);
}

/*
 * Find a string within a memory chunk, stupid style!
 * Search is done backward, as the signature will be appended at the
 * end of the stream.
 */
char *
memstr(const void *h0, size_t k, const char *n0, size_t l)
{
	ssize_t i;
	const unsigned char *h = h0;

        /* Return immediately on empty needle */
        if (!l) return (char *)h;

        /* Return immediately when needle is longer than haystack */
        if (k<l) return 0;

	for (i=k-l; i>0; i--) {
		if (memcmp(h+i, n0, l) == 0)
			return (char *)(h+i);
	}
	return NULL;
}

/*
 * read chunks of data from a stream into a buffer, and return the size of the
 * buffer
 */
static size_t
bufferize(unsigned char **buf, FILE *fp)
{
	size_t n, len = 0;
	unsigned char chunk[_POSIX_MAX_INPUT], *tmp;

	/*
	 * For each chunk read, reallocate the buffer size to fit the newly
	 * read data, and copy it over
	 */
	while (!feof(fp) && (n = fread(chunk, 1, _POSIX_MAX_INPUT, fp)) > 0) {
		if ((tmp = realloc(*buf, len + n)) == NULL) {
			free(*buf);
			*buf = NULL;
			return 0;
		}

		*buf = tmp;
		memcpy((*buf) + len, chunk, n);
		len += n;
	}

	return len;
}

/*
 * Copy the full content of the buffer, minus the signature to the given
 * pointer
 */
static size_t
extractmsg(unsigned char **msg, unsigned char *buf, size_t buflen)
{
	size_t len = 0;
	unsigned char *sig;

	if (armored) {
		/* signature start is identified by SIGBEGIN */
		sig = (unsigned char *) memstr(buf, buflen, SIGBEGIN, strlen(SIGBEGIN));
	} else {
		/* signatures are 64 bytes long, appended to the stream */
		sig = buf + buflen - 64;
	}

	/* if signature is not found, return the whole buffer */
	if (sig == NULL) {
		len = buflen;
	} else {
		len = sig - buf;
	}

	*msg = malloc(len);
	memcpy(*msg, buf, len);

	return len;
}

/*
 * Copy the signature at the end of the buffer to the given pointer
 */
static size_t
extractsig(unsigned char **sig, unsigned char *buf, size_t len)
{
	off_t i;
	size_t n, siglen = 0;
	unsigned char *begin, *end, *tmp;
	unsigned char base64[76];

	if (armored) {
		/* search start and end strings for the signatures */
		begin = (unsigned char *) memstr(buf, len, SIGBEGIN, strlen(SIGBEGIN)) + strlen(SIGBEGIN);
		end   = (unsigned char *) memstr(buf, len, SIGEND, strlen(SIGEND));
		if (!(begin && end) || end != (buf + len - strlen(SIGEND)))
			return 0;
	}

	/* ed25519 signatures are 64 bytes longs */
	*sig = malloc(64);
	if (*sig == NULL)
		return 0;

	/* zero signature memory */
	memset(*sig, 0, 64);

	if (armored) {
		/*
		 * base64 signature are wrapped at 76 chars.
		 * 76 being a multiple of 4, it means we can decode the signature in
		 * chunks of 76 bytes, and concatenate them together to get the
		 * original data.
		 */
		for (i = 0; begin+i < end; i+=77) {
			/*
			 * black magic pointer arithmetic there..
			 * if we reached the "end" pointer, it means we're at the end
			 * of the signature.
			 * The line length is either 76 bytes long, or less for the
			 * last line. In the latter case, the trailing \n must be
			 * trimmed before decoding.
			 */
			n = begin+i+76 < end ? 76 : end - (begin + i) - 1;
			memset(base64, 0, 76);
			memcpy(base64, begin+i, n);

			n = base64_decode((char **)&tmp, base64, n);
			memcpy((*sig) + siglen, tmp, n);
			siglen += n;
			free(tmp);
		}
	} else {
		/* assume the signature is the last 64 bytes of the stream */
		if (len > 64 && memcpy((*sig), buf + len - 64, 64))
			siglen = 64;
	}

	return siglen;
}

/*
 * Creates a set of ed25519 key pairs on disk.
 */
static int
createkeypair(const char *alias)
{
	size_t klen = 0;
	FILE *fp = NULL;
	char fn[PATH_MAX];
	unsigned char pub[32], priv[64], base64[96], *buf = NULL;

	crypto_sign_ed25519_keypair(pub, priv);

	memset(fn, 0, PATH_MAX);
	memset(base64, 0, sizeof(base64));

	/* write public key to "<alias>.pub" */
	snprintf(fn, sizeof(fn) - 4, "%s.pub", alias);
	if (verbose)
		fprintf(stderr, "Creating public key %s\n", fn);
	if ((fp = fopen(fn, "w")) == NULL) {
		perror(fn);
		return -1;
	}
	klen = base64_encode((char **)&buf, pub, sizeof(pub));
	memcpy(base64, buf, klen);
	base64[klen] = '\n';
	if (fwrite(base64, 1, klen+1, fp) < klen) {
		fclose(fp);
		perror(fn);
		return -1;
	}
	fclose(fp);
	free(buf);

	/* write private key to "<alias>.key" */
	snprintf(fn, sizeof(fn) - 4, "%s.key", alias);
	if (verbose)
		fprintf(stderr, "Creating private key %s\n", fn);
	if ((fp = fopen(fn, "w")) == NULL) {
		perror(fn);
		return -1;
	}
	klen = base64_encode((char **)&buf, priv, sizeof(priv));
	memcpy(base64, buf, klen);
	base64[klen] = '\n';
	if (fwrite(base64, 1, klen + 1, fp) < klen) {
		fclose(fp);
		perror(fn);
		return -1;
	}
	fclose(fp);
	free(buf);

	return 0;
}

int
readkey(FILE *fp, unsigned char **k)
{
	size_t len;
	char b64[96];
	len = fread(b64, 1, sizeof(b64), fp);
	return base64_decode((char **)k, (unsigned char *)b64, len);
}

/*
 * Buffer a data stream, sign it, and write the buffer + base64 encoded
 * signature to stdout
 */
int
sign(FILE *fp, FILE *key)
{
	unsigned long long mlen, slen;
	char *base64;
	unsigned char sig[64], *priv = NULL, *msg = NULL, *buf = NULL;

	if (key == NULL)
		return ERR_NOKEY;

	if (readkey(key, &priv) < 64)
		return ERR_NOKEY;

	mlen = bufferize(&msg, fp);
	if (mlen == 0) {
		free(priv);
		return ERR_NOMSG;
	}

	if (verbose)
		fprintf(stderr, "Signing stream (%llu bytes)\n", mlen);

	buf = malloc(mlen + 64);

	crypto_sign_ed25519(buf, &slen, msg, mlen, priv);
	free(priv);

	memcpy(sig, buf, 64);
	free(buf);

	/* write buffer to stdout .. */
	fwrite(msg, 1, mlen, stdout);
	free(msg);

	if (armored) {
		/* .. followed by the signature delimiter .. */
		fwrite(SIGBEGIN, 1, strlen(SIGBEGIN), stdout);

		/* .. then the base64 encoded signature .. */
		mlen = base64_encode(&base64, sig, 64);
		base64_fold(stdout, base64, mlen, 0);
		free(base64);

		/* .. and the final signature delimiter! */
		fwrite(SIGEND, 1, strlen(SIGEND), stdout);
	} else {
		/* simply write the signature as-is */
		fwrite(sig, 1, 64, stdout);
	}

	return 0;
}

/*
 * Check a buffer against all files in the $KEYRING directory set in the
 * environment.
 */
static int
check_keyring(unsigned char *sig, unsigned char *msg, size_t len)
{
	int ret = 0;
	unsigned long long n = 0, buflen;
	DIR *dirp = NULL;
	FILE *key = NULL;
	struct dirent *dt = NULL;
	char *keyring = NULL, path[PATH_MAX];
	unsigned char *pub;

	/* get the keyring from the environment */
	keyring = getenv("KEYRING");
	if (keyring == NULL) {
		if (verbose)
			fprintf(stderr, "KEYRING not set\n");
		return ERR_NORING;
	}

	dirp = opendir(keyring);
	if (dirp == NULL) {
		perror(keyring);
		return ERR_NORING;
	}

	/* loop through all entries in the $KEYRING directory */
        while ((dt = readdir(dirp)) != NULL) {
		/* ignore entries that are not regular files */
                if (dt->d_type != DT_REG)
			continue;

		/* ignore all entries that are not 32 bytes long */
		if (dt->d_reclen != 32)
			continue;

		/* set public key file path and store its content */
                n = strnlen(keyring, PATH_MAX);
                memset(path, 0, PATH_MAX);
                memcpy(path, keyring, n);
                path[n] = '/';
                memcpy(path+n+1, dt->d_name, dt->d_reclen);
		if ((key = fopen(path, "r")) == NULL) {
			perror(path);
			continue;
		}
		if (readkey(key, &pub) < 32) {
			perror(path);
			fclose(key);
			free(pub);
			continue;
		}

		/* check message for the given public key */
		ret += crypto_sign_ed25519_open(sig, &buflen, msg, len, pub) == -1 ? 0 : 1;
		free(pub);
		if (ret) {
			if (verbose)
				fprintf(stderr, "Key match: %s\n", path);
			break;
		}
        }

        closedir(dirp);
	return !ret;
}

/*
 * Check the given stream against the key provided. If the stream pointer
 * supposed to hold the key is NULL, check the stream against all public keys
 * located in the $KEYRING directory.
 */
static int
check(FILE *fp, FILE *key)
{
	int ret = 0;
	unsigned long long len, dummylen;
	unsigned char *pub, *sig, *msg, *buf = NULL, *dummybuf = NULL;

	if ((len = bufferize(&buf, fp)) == 0)
		return ERR_NOMSG;

	if (verbose)
		fprintf(stderr, "Extracting signature from input\n");

	int slen = extractsig(&sig, buf, len);
	if (slen != 64) {
		if (verbose)
			fprintf(stderr, "ERROR: No valid signature found (%d bytes)\n", slen);

		free(buf);
		return ERR_NOSIG;
	}

	if ((len = extractmsg(&msg, buf, len)) == 0) {
		free(buf);
		free(sig);
	}

	free(buf);

	if (verbose)
		fprintf(stderr, "Verifying stream (%llu bytes)\n", len);

	if (key) {
		if (readkey(key, &pub) != 32)
			return ERR_NOKEY;

		buf = malloc(len + 64);
		memcpy(buf, sig, 64);
		memcpy(buf + 64, msg, len);
		dummybuf = malloc(len + 64);

		ret = crypto_sign_ed25519_open(dummybuf, &dummylen, buf, len + 64, pub);
		free(buf);
		free(dummybuf);
	} else {
		ret = check_keyring(sig, msg, len);
	}

	/*
	 * if we're able to verify the signature, dump buffer's content to
	 * stdout
	 */
	if (!ret)
		fwrite(msg, 1, len, stdout);

	if (verbose)
		fprintf(stderr, "Stream check %s\n", ret ? "FAILED" : "OK");

	free(msg);
	free(sig);

	return ret;
}

/*
 * Remove a signature from a stream, and dump it to stdout
 */
static int
trimsig(FILE *fp)
{
	size_t len = 0;
	unsigned char *msg = NULL, *buf = NULL;

	len = bufferize(&buf, fp);
	if (!buf)
		return -1;

	len = extractmsg(&msg, buf, len);
	if (!msg) {
		free(buf);
		return ERR_NOMSG;
	}

	fwrite(msg, 1, len, stdout);

	free(buf);
	free(msg);

	return 0;
}

int
main(int argc, char *argv[])
{
	int ret = 0, action = ACT_CHCK;
	FILE *key = NULL, *fp = NULL;

	ARGBEGIN{
	case 'a':
		armored = 1;
		break;
	case 'f':
		key = fopen(EARGF(usage()), "r");
		break;
	case 'g':
		return createkeypair(EARGF(usage()));
		break; /* NOTREACHED */
	case 's':
		action = ACT_SIGN;
		break;
	case 't':
		action = ACT_TRIM;
		break;
	case 'v':
		verbose = 1;
		break;
	default:
		usage();
	}ARGEND;

	/* if no argument is provided, read stdin */
	fp = argc ? fopen(*argv, "r") : stdin;

	switch (action) {
	case ACT_SIGN:
		ret |= sign(fp, key);
		break;
	case ACT_CHCK:
		ret |= check(fp, key);
		break;
	case ACT_TRIM:
		ret |= trimsig(fp);
		break;
	}

	fclose(fp);
	if (key)
		fclose(key);

	return ret;
}
